/*
  (c) 2003-2007 Philipp Klaus Krause pkk@spth.de

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.
   
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
   
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <string>
#include <iostream>
#include <fstream>
#include <utility>
#include <math.h>
#include <sys/types.h>
#include <png.h>

class output
{
public:
	output(std::ostream& o);
	~output(void);
	void set_binary(bool b);
	void set_rle(bool r);
	void set_name(const std::string& n);
	void write(uint8_t b);
private:
	std::ostream *out;
	bool binary;
	bool rle;
	std::string name;
	bool rle_started;
	uint8_t rle_last;
	uint8_t rle_counter;
	bool started;
};

output::output(std::ostream& o)
{
	out = &o;
	binary = false;
	rle = false;
	name = "data";
	started = false;
}

output::~output(void)
{
	if(rle)
	{
		if(rle_started)
			(*out) << ", ";
		if(rle_counter == 1 && rle_last != 0xff)
		{
			(*out) << int(rle_last);
		}
		else if(rle_counter == 2 && rle_last != 0xff)
		{
			(*out) << int(rle_last) << ", " << int(rle_last);
		}
		else
		{
			(*out) << int(0xff) << ", " << int(rle_counter) << ", " << int(rle_last);
		}
	}
	if(!binary)
		(*out) << "};\n";
}

void output::set_binary(bool b)
{
	binary = b;
}

void output::set_rle(bool r)
{
	rle = r;
}

void output::set_name(const std::string& n)
{
	name = n;
}

void output::write(uint8_t b)
{
	if(!started)
	{
		if(!binary)
		{
			(*out) << "const unsigned char " << name << "[] = {";
			out->setf(std::ios_base::hex, std::ios_base::basefield);
			out->setf(std::ios_base::showbase);
		}
		
	}

	if(!binary)
	{	
		if(!rle)
		{
			if(started)
				(*out) << ", ";
			(*out) << int(b);
		}
		else
		{
			if(!started)
			{
				rle_counter = 1;
				rle_last = b;
				rle_started = false;
			}
			else
			{
				if(rle_last == b && rle_counter < 255)
					rle_counter++;
				else
				{
					if(rle_started)
						(*out) << ", ";
					if(rle_counter == 1 && rle_last != 0xff)
					{
						(*out) << int(rle_last);
					}
					else if(rle_counter == 2 && rle_last != 0xff)
					{
						(*out) << int(rle_last) << ", " << int(rle_last);
					}
					else
					{
						(*out) << int(0xff) << ", " << int(rle_counter) << ", " << int(rle_last);
					}
					rle_last = b;
					rle_counter = 1;
					rle_started = true;
				}
			}
		}
	}
	else
		out->put(b);

	started = true;
}

// Convert 8 RGBA pixels (one row of character).
inline void rgba2cv(const void *in, u_int8_t *pattern, u_int8_t *color)
{
  unsigned int i, j, k, l;
  float row[8][4];
  for(i = 0; i < 8; i++)  // Conversion to 9 bit color + 1 bit alpha mask.
  {
    row[i][0] = float(((u_int8_t *)in)[i * 4 + 0]) / 36.0f;
    if(row[i][0] >= 7) row[i][0] = 7;
    row[i][1] = float(((u_int8_t *)in)[i * 4 + 1]) / 36.0f;
    if(row[i][1] >= 7) row[i][1] = 7;
    row[i][2] = float(((u_int8_t *)in)[i * 4 + 2]) / 36.0f;
    if(row[i][2] >= 7) row[i][2] = 7;
    row[i][3] = ((u_int8_t *)in)[i * 4 + 3] ? 7.0f : 0.0f;
  }

  // Colecovision colors in RGBA format.
  int cvcols[16][4] = {{0, 0, 0, 0}, {0, 0, 0, 7}, {1, 6, 1, 7}, {3, 7, 3, 7}, {7, 1, 1, 7}, {7, 3, 2, 7}, {1, 1, 5, 7}, {7, 6, 2, 7}, {1, 1, 7, 7}, {3, 3, 7, 7}, {1, 6, 6, 7}, {4, 6, 6, 7}, {1, 4, 1, 7}, {5, 2, 6, 7}, {5, 5, 5, 7}, {7, 7, 7, 7}};

  double dist[16][16];

  // Find best two colors
  // Calculate differences
  for(i = 0; i < 16; i++)
	{
    for(j = 0; j < 16; j++)
    {
      dist[i][j] = 0;
      for(k = 0; k < 8; k++)
      {
        if((cvcols[i][3] != 0) && (cvcols[j][3] != 0) || (row[k][3] != 0))
        {
          dist[i][j] += std::min(fabs(cvcols[i][0] - row[k][0]), fabs(cvcols[j][0] - row[k][0]));
          dist[i][j] += std::min(fabs(cvcols[i][1] - row[k][1]), fabs(cvcols[j][1] - row[k][1]));
          dist[i][j] += std::min(fabs(cvcols[i][2] - row[k][2]), fabs(cvcols[j][2] - row[k][2]));
        }
        dist[i][j] += std::min(fabs(cvcols[i][3] - row[k][3]), fabs(cvcols[j][3] - row[k][3]));
      }
    }
	}

  for(float fd = 0.0f;; fd+= 0.1f)
  {
    for(j = 0; j < 16; j++)
      for(l = 0; l < 16; l++)
        if(dist[j][l] <= fd)  // Best match
        {
          *color = ((j << 4) + l);
          *pattern = 0;
          for(k = 0; k < 8; k++)
          {
            (*pattern) <<= 1;
            double dist1 = 0, dist2 = 0;
            dist1 += fabs(cvcols[j][0] - row[k][0]);
            dist2 += fabs(cvcols[l][0] - row[k][0]);
            dist1 += fabs(cvcols[j][1] - row[k][1]);
            dist2 += fabs(cvcols[l][1] - row[k][1]);
            dist1 += fabs(cvcols[j][2] - row[k][2]);
            dist2 += fabs(cvcols[l][2] - row[k][2]);
            if(row[k][3] != cvcols[j][3]) dist1 += 0xff;
            if(row[k][3] != cvcols[l][3]) dist1 += 0xff;
            if(dist1 <= dist2) (*pattern)++;     
          }
          return;
        }
  }
  return;
}

// Convert 8 RGB pixels (one row of character).
inline void rgb2cv(const void *in, u_int8_t *pattern, u_int8_t *color)
{
  unsigned int i, j, k, l;
  double row[8][3];
  for(i = 0; i < 8; i++)  // Conversion to 9 bit color.
  {
    row[i][0] = double(((u_int8_t *)in)[i * 3 + 0]) / 36.43;
    row[i][1] = double(((u_int8_t *)in)[i * 3 + 1]) / 36.43;
    row[i][2] = double(((u_int8_t *)in)[i * 3 + 2]) / 36.43;
  }

  // Colecovision colors in RGB format.
  double cvcols[15][3] = {{0.0f, 0.0f, 0.0f}, {1.0f, 6.0f, 1.0f}, {3.0f, 7.0f, 3.0f}, {1.0f, 1.0f, 7.0f}, {2.0f, 3.0f, 7.0f}, {5.0f, 1.0f, 1.0f}, {2.0f, 6.0f, 7.0f}, {7.0f, 1.0f, 1.0f}, {7.0f, 3.0f, 3.0f}, {6.0f, 6.0f, 1.0f}, {6.0f, 6.0f, 4.0f}, {1.0f, 4.0f, 1.0f}, {6.0f, 2.0f, 5.0f}, {5.0f, 5.0f, 5.0f}, {7.0f, 7.0f, 7.0f}};

  double dist[15][15];

  // Find best two colors
  // Calculate differences
  for(i = 0; i < 15; i++)
    for(j = 0; j < 15; j++)
    {
      dist[i][j] = 0;
      for(k = 0; k < 8; k++)
      {
        double d0 = 0.0, d1 = 0.0;
        d0 += fabs(cvcols[i][0] - row[k][0]);
        d1 += fabs(cvcols[j][0] - row[k][0]);
        d0 += fabs(cvcols[i][1] - row[k][1]) * 1.5;
        d1 += fabs(cvcols[j][1] - row[k][1]) * 1.5;
        d0 += fabs(cvcols[i][2] - row[k][2]);
        d1 += fabs(cvcols[j][2] - row[k][2]);
        dist[i][j] += std::min(d0, d1);
      }
    }
  for(float fd = 0;;fd += 0.04f)
  {
    for(j = 0; j < 15; j++)
      for(l = 0; l < 15; l++)
        if(dist[j][l] <= fd)  // Best match
        {
          *color = ((j + 1) << 4) + (l + 1);
          *pattern = 0;
          for(k = 0; k < 8; k++)
          {
            (*pattern) <<= 1;
            double dist1 = 0.0, dist2 = 0.0;
            dist1 += fabs(cvcols[j][0] - row[k][0]);
            dist2 += fabs(cvcols[l][0] - row[k][0]);
            dist1 += fabs(cvcols[j][1] - row[k][1]);
            dist2 += fabs(cvcols[l][1] - row[k][1]);
            dist1 += fabs(cvcols[j][2] - row[k][2]);
            dist2 += fabs(cvcols[l][2] - row[k][2]);
            if(dist1 <= dist2) (*pattern)++;    
          }
          return;
        }
  }
  return;
}

inline void rgba2bw(const void *in, u_int8_t *pattern)
{
}

inline void rgb2bw(const void *in, u_int8_t *pattern)
{
}

void usage(void)
{
  std::cerr
    << '\n' << PACKAGE << ' ' << PACKAGE_VERSION << " (c) 2003-2007 Philipp Krause " << PACKAGE_BUGREPORT
    << "\nUsage: png2cv [-b][--binary][--black] file\n"
    << "\n-b, --binary: binary output instead of C source."
    << "\n--monochrome, --black: Generate data for bitmap text mode instead of bitmap mode."
		<< "\n--rle: RLE encode output (obsolete)."
    << "\n--verbose: Verbose output.\n";
  std::cerr << "The conversion of images with an alpha channel is currently broken as is the --black option!\n";
}

int main(int argc, char** argv)
{
	std::string ifile, patternofile, colorofile;
	 bool binary = false;
	bool color = true;
	bool verbose = false;
	bool rle = false;

  int i;

  if(argc == 1)
  {
    std::cerr << "Too few arguments.\n";
    usage();
    return(-1);
  }

  for(i = 1; i < argc - 1; i++)
  {
    std::string stra = argv[i];
    if(stra == "-b") binary = true;
		else if(stra == "--binary") binary = true;
		else if(stra == "--monochrome") color = false;
		else if(stra == "--rle")
		{
			rle = true;
			std::cout << "The --rle option is obsolete. Use binary output with external compression program instead.\n";
		}
    else if(stra == "--verbose") verbose = true;
    else
    {
      usage();
      return(-1);
    }
  }
  
  ifile = argv[argc - 1];

  patternofile = ifile;
  colorofile = ifile;
  if(binary)
  {
    patternofile += ".pattern.bin";
    colorofile += ".color.bin"; 
  }
  else
  {
    patternofile += ".pattern.c";
    colorofile += ".color.c";
  }



  FILE *image_file;
  unsigned char signature[8]; // PNG signature
  png_structp  png_ptr;
  png_infop info_ptr;
  png_uint_32 height;
  png_uint_32 width;
  int depth;
  int colortype;
  png_uint_32	i_png, rowbytes;

  if(!(image_file = fopen(ifile.c_str(), "rb")))
  {
    std::cerr << "File not found.\n";
    usage();
    return(-1);
  }
  if(fread(signature, 1, 8, image_file) < 8)
  {
    std::cerr << "Invalid file.\n";
    usage();
    return(-1);
  }
  if(!png_check_sig(signature, 8))
  {
    std::cerr << "Not a PNG image file.\n";
    usage();
    return(-1);
  }

  // initialize libpng
  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if(!png_ptr)
  {
    std::cerr << "Out of memory.\n";
    return(-1);
  }
  info_ptr = png_create_info_struct(png_ptr);
  if(!info_ptr)
  {
    png_destroy_read_struct(&png_ptr, 0, 0);
    std::cerr << "Out of memory.\n";
    return(-1);
  }

  if(setjmp(png_ptr->jmpbuf))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, 0);
    std::cerr << "libpng error.\n";
    return(-1);
  }


  png_init_io(png_ptr, image_file);
  png_set_sig_bytes(png_ptr, 8);
  png_read_info(png_ptr, info_ptr);
  png_get_IHDR(png_ptr, info_ptr, &width, &height, &depth, &colortype, 0, 0, 0);

  // Convert everything to RGB or RGBA.
  if(colortype == PNG_COLOR_TYPE_PALETTE) png_set_palette_to_rgb(png_ptr);  // Convert indexed to RGB
  if(colortype == PNG_COLOR_TYPE_GRAY && depth < 8) png_set_gray_1_2_4_to_8(png_ptr); // Convert 1, 2 and 4 bit greyscale to 8 bit.
  if(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) png_set_tRNS_to_alpha(png_ptr); // Convert bitmask tranparency to full Alphachannel.
  if(depth == 16) png_set_strip_16(png_ptr);  // Reduce color depth of 48 and 64 bit images.
  if(colortype == PNG_COLOR_TYPE_GRAY || colortype == PNG_COLOR_TYPE_GRAY_ALPHA) png_set_gray_to_rgb(png_ptr);  // Convert greyscale to RGB.
  
  png_read_update_info(png_ptr, info_ptr);

  png_bytep *Zeilenzeiger;
  unsigned char *image_data;
  try
  {
    try
    {
      Zeilenzeiger = new png_bytep[height];
      rowbytes = png_get_rowbytes(png_ptr, info_ptr);    
      try
      {
        image_data = new unsigned char[rowbytes * height];
      }
      catch(...)
      {
        delete[] Zeilenzeiger;
        throw;
      }
    }
    catch(...)
    {
      png_destroy_read_struct(&png_ptr, &info_ptr, 0);
      throw;
    }
  }
  catch(std::bad_alloc)
  {
    std::cerr << "Out of memory.";
    return(-1);
  }
  catch(...)
  {
    std::cerr << "Unknown error";
    return(-1);
  }

  for(i_png = 0; i_png < height; ++i_png) Zeilenzeiger[i_png] = image_data + i_png * rowbytes;
  png_read_image(png_ptr, Zeilenzeiger);
  png_read_end(png_ptr, 0);
  fclose(image_file);
  delete[] Zeilenzeiger;

  if(png_ptr && info_ptr)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, 0);
    png_ptr = 0; info_ptr = 0;
  }

  if((width % 8) || (height % 8))
  {
    delete[] image_data;
    std::cerr << "Width and height must be multiples of 8.";
    return(-1);
  }

  u_int8_t cvimage;
  u_int8_t cvcolor;
  std::ofstream patterno;
  std::ofstream coloro;

  patterno.open(patternofile.c_str(), std::ios::out | std::ios::trunc);
  if(color) coloro.open(colorofile.c_str(), std::ios::out | std::ios::trunc);

	output *patternout;
	output *colorout;
	{
		patternout = new output(patterno);
		patternout->set_name("pattern");
		patternout->set_binary(binary);
		patternout->set_rle(rle);
	}
	if(color)
	{
		colorout = new output(coloro);
		colorout->set_name("color");
		colorout->set_binary(binary);
		colorout->set_rle(rle);
	}
  if(verbose)
  {
    std::cout << "Converting " << width << "x" << height << " " << ((colortype == PNG_COLOR_TYPE_RGBA || colortype == PNG_COLOR_TYPE_GRAY_ALPHA) ? "RGBA" : "RGB") << " image.\n";
    std::cout << "A . for every Colecovision character line written.\n";
    std::cout.flush();
  }
	int j, k;
	for(i = 0; i < int(height) / 8; i++)
		for(j = 0; j < int(width) / 8; j++)
		{
			for(k = 0; k < 8; k++)
			{     
				if(color)
				{
					if(colortype == PNG_COLOR_TYPE_RGBA || colortype == PNG_COLOR_TYPE_GRAY_ALPHA) // RGBA
						rgba2cv(&(image_data[(i * 8 * width + j * 8 + k * width) * 4]), &cvimage, &cvcolor);
					else  // RGB
						rgb2cv(&(image_data[(i * 8 * width + j * 8 + k * width) * 3]), &cvimage, &cvcolor);
				}
				else
 				{
					if(colortype == PNG_COLOR_TYPE_RGBA || colortype == PNG_COLOR_TYPE_GRAY_ALPHA) // RGBA
						rgba2bw(&(image_data[(i * 8 * width + j * 8 + k * width) * 4]), &cvimage);
 					else  // RGB
						rgb2bw(&(image_data[(i * 8 * width + j * 8 + k * width) * 3]), &cvimage);
				}

				patternout->write(cvimage);
				if(color)
					colorout->write(cvcolor);
			}
			if(verbose)
			{
				std::cout << '.';
				std::cout.flush();
			}
		}

	delete patternout;
	if(color)
		delete colorout;
  patterno.close();
  if(color) coloro.close();
  if(verbose) std::cout << "\ndone.\n";

  delete[] image_data;

  return(0);
}
